"""Newton heavy lifting functions and helpers
"""

import jax
import time
import jax.numpy as jnp
from grgrjax import callback_func, amax, newton_jax_jit


def iteration_step(carry):
    (y, dampening, cnt), (x, f, jvp_func, jacobian, factor), (_, tol, maxit) = carry
    _, v = jvp_func(x, y)
    v = jax.scipy.linalg.lu_solve(jacobian, v)
    dampening = jnp.minimum(dampening, factor*jnp.abs((y.T@y)/(v.T@y)))
    diff = f-v
    y += dampening*diff
    eps = amax(diff)
    return (y, dampening, cnt+1), (x, f, jvp_func, jacobian, factor), (eps, tol, maxit)


def iteration_cond(carry):
    (_, _, cnt), _, (eps, tol, maxit) = carry
    return jnp.logical_and(cnt <= maxit, eps > tol)


def jvp_while_body(carry):
    (x, _, _, cnt), (jvp_func, jacobian, maxit,
                     nsteps, tol, factor, verbose) = carry
    # first iteration
    f, _ = jvp_func(x, jnp.zeros_like(x))
    f = jax.scipy.linalg.lu_solve(jacobian, f)
    # other iterations
    iteration_tol = jnp.minimum(1e-5, 1e-1*amax(f))
    init = ((f, 1., 0), (x, f, jvp_func, jacobian, factor),
            (1e8, iteration_tol, nsteps))
    (y, dampening, cnt_inner), _, _ = jax.lax.while_loop(
        iteration_cond, iteration_step, init)
    return (x-y, amax(f), dampening, cnt+cnt_inner), (jvp_func, jacobian, maxit, nsteps, tol, factor, verbose)


def jvp_while_cond(carry):
    (_, err, dampening, cnt), (_, _, maxit, nsteps, tol, _, verbose) = carry
    cond = jnp.logical_and(err > tol, cnt < maxit)
    verbose = jnp.logical_and(cond, verbose)
    verbose = jnp.logical_and(cnt, verbose)
    jax.debug.callback(callback_func, cnt, err, dampening, verbose=verbose)
    return cond


def sweep_banded_down(val, i):
    jav_func, fmod, forward_mat, X, shocks = val
    # calculate value and jacobians
    fval, (jac_f2xLag, jac_f2x, jac_f2xPrime) = jav_func(
        X[i], X[i+1], X[i+2], shocks=shocks[i])
    bmat = jac_f2x - jac_f2xLag @ forward_mat
    forward_mat = jnp.linalg.solve(bmat, jac_f2xPrime)
    fmod = jnp.linalg.solve(bmat, fval - jac_f2xLag @ fmod)
    return (jav_func, fmod, forward_mat, X, shocks), (fmod, forward_mat)


def sweep_banded_up(val, i):
    forward_mat, fvals, fval = val
    # go backwards in time
    fval = fvals[i] - forward_mat[i] @ fval
    return (forward_mat, fvals, fval), fval


def check_status(err, cnt, maxit, tol):
    """Check whether to exit iteration and compile error message"""
    # exit causes
    if err < tol:
        return True, (True, "The solution converged.")
    elif jnp.isnan(err):
        return True, (False, "Function returns 'NaN's.")
    elif cnt > maxit:
        return True, (False, f"Maximum number of {maxit} iterations reached.")
    else:
        return False, (False, "")


def newton_for_jvp(jvp_func, jacobian, x_init, verbose, tol=1e-8, maxit=500, nsteps=30, factor=1.5):

    start_time = time.time()
    x = x_init[1:-1].flatten()

    (x, err, dampening, cnt), _ = jax.lax.while_loop(jvp_while_cond, jvp_while_body,
                                                     ((x, 1., 0., 0), (jvp_func, jacobian, maxit, nsteps, tol, factor, verbose)))
    ltime = time.time() - start_time
    callback_func(cnt, err, dampening, ltime, verbose)
    _, (success, mess) = check_status(err, cnt, maxit, tol)
    # compile error/report message
    if not success and not jnp.isnan(err):
        mess += f" Max. error is {err:1.2e}."

    return x, not success, mess


def newton_for_banded_jac(jav_func, nvars, horizon, X, shocks, verbose, maxit=30, tol=1e-8):

    st = time.time()
    cnt = 0

    while True:

        _, (fvals, forward_mat) = jax.lax.scan(sweep_banded_down, (jav_func, jnp.zeros(
            nvars), jnp.zeros((nvars, nvars)), X, shocks), jnp.arange(horizon-1))
        _, out = jax.lax.scan(sweep_banded_up, (forward_mat, fvals, jnp.zeros(
            nvars)), jnp.arange(horizon-1), reverse=True)

        X = X.at[1:-1].add(-out)
        err = amax(out)
        cnt += 1

        ltime = time.time() - st

        if verbose:
            info_str = f'    Iteration {cnt:3d} | max error {err:.2e} | lapsed {ltime:3.4f}s'
            print(info_str)

        do_break, (success, mess) = check_status(err, cnt, maxit, tol)
        if do_break:
            break

    if not success:
        mess += f" Max. error is {err:1.2e}."

    return X, False, ''


def newton_jax_jit_wrapper(func, init, **args):
    """Wrapper around grgrjax.newton.newton_jax_jit to return flag and message
    """

    if 'tol' not in args:
        args['tol'] = 1e-8
    if 'maxit' not in args:
        args['maxit'] = 30

    x, (f, _), cnt, flag = newton_jax_jit(func, init, **args)
    err = amax(f)
    _, (success, mess) = check_status(err, cnt, args['maxit'], args['tol'])
    flag |= not success

    # compile error/report message
    if not success and not jnp.isnan(err):
        mess += f" Max. error is {err:1.2e}."
    return x, flag, mess



# -*- coding: utf-8 -*-

import jax
import time
import jax.numpy as jnp
from ..utilities.jacobian import get_stst_jacobian
from ..parser.checks import check_if_compiled, write_compiled_objects
from ..parser.build_functions import build_aggr_het_agent_funcs, get_stst_derivatives


def find_path_linear(model, shock=None, init_state=None, parameters=None, horizon=200, verbose=True):
    """Find the linear expected trajectory given an initial state.

    Parameters
    ----------
    model : PizzaModel
        PizzaModel instance
    init_state : array
        initial state
    parameters : dict, optional
        alternative parameters. Warning: do only change those parameters that are invariant to the steady state.
    shock : tuple, optional
        shock in period 0 as in `(shock_name_as_str, shock_size)`. NOTE: Not (yet) implemented.
    horizon : int, optional
        number of periods to simulate
    verbose : bool, optional
        degree of verbosity. 0/`False` is silent

    Returns
    -------
    x : array
        array of the trajectory
    flag : bool
        for consistency. Always returns `True`
    """

    if shock is not None:
        raise NotImplementedError(
            "Shocks are not (yet) implemented for the linear solution.")

    if not model.get('distributions'):
        raise NotImplementedError(
            "Models without heterogeneous agents are not (yet) implemented for the linear solution.")

    st = time.time()

    # get model variables
    stst = jnp.array(list(model['stst'].values()))
    nvars = len(model["variables"])
    pars = jnp.array(
        list((parameters if parameters is not None else model["parameters"]).values()))
    shocks = model.get("shocks") or ()
    x_stst = jnp.ones((horizon + 1, nvars)) * stst

    # deal with shocks
    zero_shocks = jnp.zeros((horizon-1, len(shocks))).T

    x0 = jnp.array(list(init_state)) if init_state is not None else stst

    if not check_if_compiled(model, horizon, pars):
        # get derivatives via AD and compile functions
        build_aggr_het_agent_funcs(
            model, nvars, pars, stst, zero_shocks, horizon)
        derivatives = get_stst_derivatives(
            model, nvars, pars, stst, x_stst, zero_shocks, horizon, verbose)

        # accumulate steady stat jacobian
        get_stst_jacobian(model, derivatives, horizon, nvars, verbose)
        write_compiled_objects(model, horizon, pars)

    jacobian = model['jac']

    x0 -= stst
    x = - \
        jax.scipy.linalg.solve(
            jacobian[nvars:, nvars:], jacobian[nvars:, :nvars] @ x0)
    x = jnp.hstack((x0, x)).reshape(-1, nvars) + stst

    if verbose:
        duration = time.time() - st
        print(
            f"(find_path_linear:) Linear solution done ({duration:1.3f}s).")

    return x, True




"""Dynamically build functions."""

import jax
import time
import jax.numpy as jnp
from grgrjax import jvp_vmap, vjp_vmap
from .het_agent_funcs import *


def func_stst_rep_agent(y, func_pre_stst, func_eqns):
    x, par = func_pre_stst(y)
    x = x[..., None]
    return func_eqns(x, x, x, x, pars=par), None


def func_stst_het_agent(y, func_pre_stst, find_stat_vf, func_stst_dist, func_eqns):

    x, par = func_pre_stst(y)
    x = x[..., None]

    vf, decisions_output, exog_grid_vars, cnt_backw = find_stat_vf(
        x, par)
    dist, cnt_forw = func_stst_dist(decisions_output)

    # TODO: for more than one dist this should be a loop...
    decisions_output_array = decisions_output[..., None]
    dist_array = dist[..., None]

    aux = (vf, decisions_output, exog_grid_vars,
           cnt_backw), (dist, cnt_forw)
    out = func_eqns(x, x, x, x, pars=par, distributions=dist_array,
                    decisions_outputs=decisions_output_array)

    return out, aux


def get_func_stst_raw(func_pre_stst, func_backw, func_stst_dist, func_eqns, shocks, init_vf, decisions_output_init, exog_grid_vars_init, tol_backw, maxit_backw, tol_forw, maxit_forw):
    """Get a function that evaluates the steady state
    """

    zshock = jnp.zeros(len(shocks))
    partial_pre_stst = jax.tree_util.Partial(func_pre_stst)
    partial_eqns = jax.tree_util.Partial(func_eqns, shocks=zshock)

    if not func_stst_dist:
        return jax.tree_util.Partial(func_stst_rep_agent, func_pre_stst=partial_pre_stst, func_eqns=partial_eqns)

    partial_backw = jax.tree_util.Partial(func_backw, shocks=zshock)
    carry = (init_vf, decisions_output_init, exog_grid_vars_init), (init_vf +
                                                                    1, 0), (partial_backw, tol_backw, maxit_backw)
    backwards_stst = jax.tree_util.Partial(backwards_sweep_stst, carry=carry)
    forwards_stst = jax.tree_util.Partial(
        func_stst_dist, tol=tol_forw, maxit=maxit_forw)

    return jax.tree_util.Partial(func_stst_het_agent, func_pre_stst=partial_pre_stst, find_stat_vf=backwards_stst, func_stst_dist=forwards_stst, func_eqns=partial_eqns)


def get_stst_derivatives(model, nvars, pars, stst, x_stst, zshocks, horizon, verbose):

    st = time.time()

    func_eqns = model['context']["func_eqns"]
    backwards_sweep = model['context']['backwards_sweep']
    second_sweep = model['context']['second_sweep']

    distSS = jnp.array(model['steady_state']['distributions'])
    decisions_outputSS = jnp.array(
        list(model['steady_state']['decisions'].values()))[..., None]

    # basis for steady state jacobian construction
    basis = jnp.zeros((nvars*(horizon-1), nvars))
    basis = basis.at[-nvars:].set(jnp.eye(nvars))

    # get steady state jacobians for dist transition
    doSS, do2x = jvp_vmap(backwards_sweep)(
        (x_stst[1:-1].flatten(), stst, zshocks), (basis,))
    _, (f2do,) = vjp_vmap(second_sweep, argnums=1)(
        (x_stst[1:-1].flatten(), doSS, stst, distSS, zshocks), basis.T)
    f2do = jnp.moveaxis(f2do, -1, 1)

    # get steady state jacobians for direct effects x on f
    jacrev_func_eqns = jax.jacrev(func_eqns, argnums=(0, 1, 2))
    f2X = jacrev_func_eqns(stst[:, None], stst[:, None], stst[:, None],
                           stst, zshocks[:, 0], pars, distSS[..., None], decisions_outputSS)

    if verbose:
        duration = time.time() - st
        print(
            f"(get_derivatives:) Derivatives calculation done ({duration:1.3f}s).")

    return f2X, f2do, do2x


def get_stacked_func_het_agents(pars, func_backw, func_dist, func_eqns, stst, vfSS, horizon, nvars):
    """Get a function that returns the (flattend) value and Jacobian of the stacked aggregate model equations.
    """

    nshpe = (nvars, horizon-1)
    # build partials of input functions
    partial_backw = jax.tree_util.Partial(func_backw, XSS=stst, pars=pars)
    partial_dist = jax.tree_util.Partial(func_dist)

    # build partials of output functions
    backwards_sweep_local = jax.tree_util.Partial(
        backwards_sweep, stst=stst, vfSS=vfSS, horizon=horizon, func_backw=partial_backw)
    forwards_sweep_local = jax.tree_util.Partial(
        forwards_sweep, horizon=horizon, func_dist=partial_dist)
    final_step_local = jax.tree_util.Partial(
        final_step, stst=stst, horizon=horizon, nshpe=nshpe, pars=pars, func_eqns=func_eqns)
    second_sweep_local = jax.tree_util.Partial(
        second_sweep, forwards_sweep=forwards_sweep_local, final_step=final_step_local)
    stacked_func_dist_local = jax.tree_util.Partial(
        stacked_func_het_agents, backwards_sweep=backwards_sweep_local, second_sweep=second_sweep_local)

    return stacked_func_dist_local, backwards_sweep_local, forwards_sweep_local, second_sweep_local


def build_aggr_het_agent_funcs(model, nvars, pars, stst, zshocks, horizon):

    shocks = model.get("shocks") or ()
    # get functions
    func_eqns = model['context']["func_eqns"]
    func_backw = model['context'].get('func_backw')
    func_dist = model['context'].get('func_dist')

    # get stuff for het-agent models
    vfSS = model['steady_state'].get('value_functions')
    distSS = jnp.array(model['steady_state']['distributions'])

    # get actual functions
    func_raw, backwards_sweep, forwards_sweep, second_sweep = get_stacked_func_het_agents(
        pars, func_backw, func_dist, func_eqns, stst, vfSS, horizon, nvars)

    # store everything
    model['context']['func_raw'] = func_raw
    model['context']['backwards_sweep'] = backwards_sweep
    model['context']['forwards_sweep'] = forwards_sweep
    model['context']['second_sweep'] = second_sweep
    model['context']['jvp_func'] = lambda primals, tangens, x0, dist0, shocks: jax.jvp(
        func_raw, (primals, x0, dist0, shocks), (tangens, jnp.zeros(nvars), jnp.zeros_like(distSS), zshocks))
    model['context']['vjp_func'] = lambda primals, tangens, x0, dist0, shocks: jax.vjp(
        lambda x: func_raw(x, x0, dist0, shocks), primals)[1](tangens)[0]



"""Checking functions for parsing.py.
"""

import sys
import jax.numpy as jnp


def check_if_defined(evars, eqns, skipped_vars):
    """Check if all variables are defined in period t.
    """
    skipped_vars = [] if skipped_vars is None else skipped_vars

    for v in evars:
        v_in_eqns = [
            v in e.replace(v + "SS", "").replace(v + "Lag",
                                                 "").replace(v + "Prime", "")
            for e in eqns
        ]
        if not any(v_in_eqns) and not v in skipped_vars:
            raise Exception(f"Variable `{v}` is not defined for time t.")
    return


def check_dublicates(evars):
    """Check if there are dublicates
    """
    dubs = [x for i, x in enumerate(evars) if x in evars[:i]]
    dubmess = (
        " The variables list contains dublicate(s): %s" % ", ".join(
            dubs) if dubs else ""
    )
    if dubs:
        print("(parse:) Warning%s" % dubmess)
    return


def check_determinancy(evars, eqns):
    """Check if the numbers of eqns/vars match.
    """
    sorted_evars = evars[:] = sorted(list(set(evars)), key=str.lower)
    if len(sorted_evars) != len(eqns):
        raise Exception(
            f"Model has {len(sorted_evars)} variables but {len(eqns)} equations."
        )
    return sorted_evars


def check_initial_values(model, shocks, par):

    init_mixed = jnp.array(list(model['init'].values()))
    init, par = model['context']['func_pre_stst'](init_mixed)
    init = init[..., None]

    # collect some information needed later
    model['init_run'] = {}

    mess = ''
    if model.get('decisions'):
        # make a test backward and forward run
        init_vf = model['init_vf']
        try:
            _, decisions_output_init, exog_grid_vars_init = model['context']['func_backw'](
                init, init, init, init, init_vf, jnp.zeros(len(shocks)), par)
        except ValueError as e:
            if str(e) == "All input arrays must have the same shape.":
                raise type(e)("Each output of the decisions stage must have the same shape as the distribution.").with_traceback(
                    sys.exc_info()[2])
            else:
                raise

        dists_init, _ = model['context']['func_stst_dist'](
            decisions_output_init, 1e-8, 1)

        if jnp.isnan(decisions_output_init).any():
            mess = 'Outputs of decision stage contains `NaN`s'
        elif jnp.isinf(decisions_output_init).any():
            mess = 'Outputs of decision stage are not finite'
        elif jnp.isnan(dists_init).any():
            mess = 'Distribution contains `NaN`s'
        elif jnp.isinf(dists_init).any():
            mess = 'Distribution is not finite'
    else:
        decisions_output_init = dists_init = exog_grid_vars_init = []

    model['init_run']['decisions_output'] = decisions_output_init
    model['init_run']['dists'] = dists_init
    model['init_run']['exog_grid_vars'] = exog_grid_vars_init

    # final test of main function
    test = model['context']['func_eqns'](init, init, init, init, jnp.zeros(len(shocks)), par, jnp.array(
        dists_init)[..., None], jnp.array(decisions_output_init)[..., None])

    if mess:
        pass
    elif jnp.isnan(test).any():
        mess += 'Output of final stage contains `NaN`s'
    elif jnp.isinf(test).any():
        mess += 'Output of final stage is not finite'

    if mess:
        raise Exception(mess + ' for initial values.')
    return


def check_shapes(distributions, init_decisions, dist_names):
    decisions_shape = init_decisions.shape
    dist_shape = tuple([d.get('n')
                       for d in distributions[dist_names[0]].values()])
    check = [(ds == decisions_shape[-len(dist_shape):][i] or ds is None)
             for i, ds in enumerate(dist_shape)]
    if not all(check):
        raise Exception(
            f"Initial decisions and the distribution have different shapes in last dimensions: {decisions_shape}, {dist_shape}")
    return


def check_if_compiled(model, horizon, pars):
    try:
        assert model['compiled_objects']['compiled_model_flag']
        assert model['compiled_objects']['horizon'] == horizon
        return jnp.allclose(model['compiled_objects']['pars'], pars)
    except:
        return False


def write_compiled_objects(model, horizon, pars):
    model['compiled_objects']['compiled_model_flag'] = True
    model['compiled_objects']['horizon'] = horizon
    model['compiled_objects']['pars'] = pars
    return



# -*- coding: utf-8 -*-

"""functions for the one-asset HANK model with labor choice. Heavily inspired by https://github.com/shade-econ/sequence-jacobian/#sequence-space-jacobian
"""

import jax
import jax.numpy as jnp
from grgrjax import jax_print, amax
from econpizza.utilities.interp import interpolate


def hh_init(a_grid, we, R, sigma_c, T):
    """The initialization for the value function
    """

    coh = R * a_grid[None, :] + \
        we[:, None] + T[:, None]
    Va = R * (0.1 * coh) ** (-sigma_c)
    return Va


def hh(Va_p, a_grid, we, trans, R, beta, sigma_c, sigma_l, vphi):
    """A single backward step via EGM
    """

    # MUC as implied by next periods value function
    uc_nextgrid = beta * Va_p
    # back out consumption and labor supply from MUC
    c_nextgrid, n_nextgrid = cn(
        uc_nextgrid, we[:, None], sigma_c, sigma_l, vphi)

    # get consumption and labor supply in grid space
    lhs = c_nextgrid - we[:, None] * n_nextgrid + \
        a_grid[None, :] - trans[:, None]
    rhs = R * a_grid

    c = interpolate(lhs, rhs, c_nextgrid)
    n = interpolate(lhs, rhs, n_nextgrid)

    # get todays distribution of assets
    a = rhs + we[:, None] * n + trans[:, None] - c
    # fix consumption and labor for constrained households
    c, n = jnp.where(a < a_grid[0], solve_cn(
        we[:, None], rhs + trans[:, None] - a_grid[0], sigma_c, sigma_l, vphi, Va_p), jnp.array((c, n)))
    a = jnp.where(a > a_grid[0], a, a_grid[0])

    # calculate new MUC
    Va = R * c ** (-sigma_c)

    return Va, a, c, n


def cn(uc, w, sigma_c, sigma_l, vphi):
    """Return optimal c, n as function of u'(c) given parameters
    """
    return jnp.array((uc ** (-1/sigma_c), (w * uc / vphi) ** (1/sigma_l)))


def solve_cn(w, trans, sigma_c, sigma_l, vphi, uc_seed):
    uc = solve_uc(w, trans, sigma_c, sigma_l, vphi, uc_seed)
    return cn(uc, w, sigma_c, sigma_l, vphi)


def solve_uc_cond(carry):
    ne, _, _ = carry
    return amax(ne) > 1e-8


def solve_uc_body(carry):
    ne, log_uc, pars = carry
    ne, ne_p = netexp(log_uc, *pars)
    log_uc -= ne / ne_p
    return ne, log_uc, pars


def solve_uc(w, trans, sigma_c, sigma_l, vphi, uc_seed):
    """Solve for optimal uc given in log uc space.
    max_{c, n} c**(1-sigma_c) + vphi*n**(1+sigma_l) s.t. c = w*n + T
    """
    log_uc = jnp.log(uc_seed)
    pars = w, trans, sigma_c, sigma_l, vphi
    _, log_uc, _ = jax.lax.while_loop(
        solve_uc_cond, solve_uc_body, (uc_seed, log_uc, pars))
    return jnp.exp(log_uc)


def netexp(log_uc, w, trans, sigma_c, sigma_l, vphi):
    """Return net expenditure as a function of log uc and its derivative
    """
    c, n = cn(jnp.exp(log_uc), w, sigma_c, sigma_l, vphi)
    ne = c - w * n - trans

    # c and n have elasticities of -1/sigma_c and 1/sigma_l wrt log u'(c)
    c_loguc = -1/sigma_c * c
    n_loguc = 1/sigma_l * n
    netexp_loguc = c_loguc - w * n_loguc

    return ne, netexp_loguc


def transfers(pi_e, Div, Tax, e_grid):
    # hardwired incidence rules are proportional to skill; scale does not matter
    tax_rule, div_rule = e_grid, e_grid
    div = Div / jnp.sum(pi_e * div_rule) * div_rule
    tax = Tax / jnp.sum(pi_e * tax_rule) * tax_rule
    T = div - tax
    return T


def wages(w, e_grid):
    we = w * e_grid
    return we


def labor_supply(n, e_grid):
    ne = e_grid[:, None] * n
    return ne



# -*- coding: utf-8 -*-

"""Functions for model parsing yaml into a working model instance. Involves a lot of dynamic function definition...
"""

from copy import copy
import yaml
import re
import os
import sys
import tempfile
import jax
import jaxlib
import jax.numpy as jnp
import importlib.util as iu
from copy import deepcopy
from inspect import getmembers, isfunction
from jax.experimental.host_callback import id_print as jax_print
from .utilities import grids, dists, interp
from .parser.compile_functions import *
from .parser.checks import *

# initialize model cache
cached_mdicts = ()
cached_models = ()


def _load_as_module(path, add_to_path=True):
    """load a file as a module
    """

    if add_to_path:
        directory = os.path.dirname(path)
        sys.path.append(directory)

    modname = os.path.splitext(os.path.basename(path))[0]
    spec = iu.spec_from_file_location(modname, path)
    module = iu.module_from_spec(spec)
    spec.loader.exec_module(module)

    return module


def parse(mfile):
    """Parse model dictionary from yaml file. This can be desirable if values should be exchanged before loading the model.

    Parameters
    ----------
    mfile : string
        path to a yaml file to be parsed

    Returns
    -------
    mdict : dict
        the parsed yaml as a dictionary
    """

    f = open(mfile)
    mtxt = f.read()
    f.close()

    mtxt = mtxt.replace("^", "**")
    mtxt = re.sub(r"@ ?\n", " ", mtxt)
    # try to detect if `~` wants to be a `-`
    mtxt = mtxt.replace("\n ~ ", "\n - ")
    mtxt = mtxt.replace("\n  ~ ", "\n  - ")
    mtxt = mtxt.replace("   ~ ", "   - ")

    # get dict
    mdict = yaml.safe_load(mtxt)
    # create nice shortcuts
    mdict['path'] = mfile
    mdict["pars"] = mdict.get("parameters")
    mdict["vars"] = mdict["variables"]

    return mdict


def _eval_strs(vdict, context={}):
    """Evaluate a dictionary of strings into a given context
    """

    if vdict is None:
        return None
    else:
        vdict = vdict.copy()

    for v in vdict:
        if isinstance(vdict[v], str):
            context[v] = eval(vdict[v], context)
            vdict[v] = eval(v, context)
        else:
            context[v] = vdict[v]

    return vdict


def _parse_external_functions_file(model):
    """Parse the functions file.
    """
    try:
        # prepare path
        if not os.path.isabs(model["functions_file"]):
            yaml_dir = os.path.dirname(model["path"])
            model["functions_file"] = os.path.join(
                yaml_dir, model["functions_file"])

        # store content
        f = open(model["functions_file"])
        model['functions_file_plain'] = f.read()
        f.close()

    except KeyError:
        pass

    return


def _initialize_context():
    context = globals().copy()
    default_funcs = {'log': jnp.log,
                     'exp': jnp.exp,
                     'sqrt': jnp.sqrt,
                     'max': jnp.maximum,
                     'min': jnp.minimum}

    context.update(default_funcs)
    return context


def _load_external_functions_file(model, context):
    """Load the functions file as a module.
    """

    try:
        # load as a module
        module = _load_as_module(model["functions_file"])

        def func_or_compiled(func): return isinstance(
            func, jaxlib.xla_extension.CompiledFunction) or isfunction(func)
        for m in getmembers(module, func_or_compiled):
            context[m[0]] = m[1]

    except KeyError:
        pass

    return False


def _compile_init_values(evars, pars, initvals, stst):
    """Combine all available information in initial guesses.
    """

    ufixed_vars = {v: .95 for v in evars if v not in stst}
    ufixed_pars = {v: .95 for v in pars if v not in stst}

    # get inital values to test the function
    init = {**ufixed_vars, **ufixed_pars}

    # structure: aggregate values first, then values of decisions functions
    if initvals is not None:
        for v in init:
            # assign aggregate values
            if v in initvals:
                init[v] = initvals[v]

    return init


def _define_subdict_if_absent(parent, sub):
    try:
        return parent[sub]
    except KeyError:
        parent[sub] = {}
        return parent[sub]


def _define_function(func_str, context):
    """Define functions from string. Writes the function into a temporary file in order to get meaningful debug traces.
    """

    # use a termporary file to get nice debug traces if things go wrong
    tmpf = tempfile.NamedTemporaryFile(mode="w", delete=False)
    tmpf.write(func_str)
    tmpf.close()

    # define the function
    exec(compile(open(tmpf.name).read(), tmpf.name, "exec"), context)

    return tmpf.name


def _get_exog_grid_var_names(distributions):
    """WIP. So far unused.
    """
    # NOTE: this will be important when implementing that grid parameters are endogenous variables
    # NOTE: when activated, backward calls already return exogenous grid vars (exog_grid_var). They are not yet stacked, and not yet an input to forward calls

    exog_grid_var_names = ()

    if False:
        # if distributions:
        for dist_name in distributions:

            dist = distributions[dist_name]
            for v in dist:
                if dist[v]['type'] in ('exogenous', 'custom_exogenous'):
                    exog_grid_var_names += tuple(dist[v]['grid_variables'])

    return exog_grid_var_names


def load(
    model,
    raise_errors=True,
    verbose=True
):
    """Load model from dict or yaml file.

    Parameters
    ----------
    model : dict or string
        either a dictionary or the path to a yaml file to be parsed
    raise_errors : bool, optional
        whether to raise errors while checking. False will let the model fail siliently for debugging. Defaults to True
    verbose : bool, optional
        inform that parsing is done. Defaults to True

    Returns
    -------
    model : PizzaModel
        The parsed model
    """

    from .__init__ import PizzaModel

    global cached_mdicts, cached_models

    if isinstance(model, str):
        full_path = model
        model = parse(model)
        model['path'] = full_path

    model = PizzaModel(model)

    # load file with additional functions as module (if it exists)
    _parse_external_functions_file(model)
    mdict_raw = deepcopy(model)

    # check if model is already cached
    if model in cached_mdicts:
        model = cached_models[cached_mdicts.index(model)]
        print("(parse:) Loading cached model.")
        return model

    model['context'] = _initialize_context()
    model['compiled_objects'] = {}
    _load_external_functions_file(model, model['context'])

    defs = model.get("definitions")
    defs = '' if defs is None else defs
    defs = '\n'.join(defs) if isinstance(defs, list) else defs
    exec(defs, model['context'])

    eqns = model["equations"].copy()

    # check if there are dublicate variables
    check_dublicates(model["variables"])
    check_dublicates(model["parameters"])
    evars = check_determinancy(model["variables"], eqns)
    # check if each variable is defined in time t (only defining xSS does not give a valid root)
    check_if_defined(evars, eqns, model.get('skip_check_if_defined'))

    # create fixed (time invariant) grids
    grids.create_grids(model.get('distributions'), model["context"])

    shocks = model.get("shocks") or ()
    _ = _define_subdict_if_absent(model, "steady_state")
    fixed_values = _define_subdict_if_absent(
        model["steady_state"], "fixed_values")
    stst = _eval_strs(fixed_values, context=model['context'])
    model["steady_state"]['fixed_evalued'] = stst

    par = _define_subdict_if_absent(model, "parameters")
    if isinstance(par, dict):
        raise TypeError(f'parameters must be a list and not {type(par)}.')
    model["parameters"] = par

    # initialize storage for all function strings
    tmpf_names = ()
    model['func_strings'] = {}

    # NOTE: currently disabled
    exog_grid_var_names = _get_exog_grid_var_names(model.get('distributions'))

    # get function strings for decisions and distributions, if they exist
    if model.get('decisions'):
        decisions_outputs = model['decisions']['outputs']
        decisions_inputs = model['decisions']['inputs']
        model['func_strings']["func_backw"] = compile_backw_func_str(
            evars, par, shocks, decisions_inputs, decisions_outputs, model['decisions']['calls'], exog_grid_var_names)
        tmpf_names += _define_function(model['func_strings']
                                       ['func_backw'], model['context']),
    else:
        decisions_outputs = []
        decisions_inputs = []

    if model.get('distributions'):
        dist_names = list(model['distributions'].keys())
        func_stst_dist_str, func_dist_str = compile_func_dist_str(
            model['distributions'], decisions_outputs)
        # store both strings
        model['func_strings']["func_stst_dist"] = func_stst_dist_str
        model['func_strings']["func_dist"] = func_dist_str
        # execute them
        tmpf_names += _define_function(func_stst_dist_str, model['context']),
        tmpf_names += _define_function(func_dist_str, model['context']),
    else:
        dist_names = []

    # collect initial guesses
    init_guesses = _eval_strs(model["steady_state"].get(
        "init_guesses"), context=model['context'])
    model["init"] = init = _compile_init_values(evars, par, init_guesses, stst)

    # get strings that contains the function definitions
    model['func_strings']["func_pre_stst"] = compile_stst_func_str(
        evars, par, stst, init)
    model['func_strings']["func_eqns"] = compile_eqn_func_str(evars, deepcopy(eqns), par, eqns_aux=model.get(
        'aux_equations'), shocks=shocks, distributions=dist_names, decisions_outputs=decisions_outputs)

    # test if model works. Writing to tempfiles helps to get nice debug traces if not
    tmpf_names += _define_function(model['func_strings']
                                   ["func_eqns"], model['context']),
    tmpf_names += _define_function(model['func_strings']
                                   ['func_pre_stst'], model['context']),

    # get the initial decision functions
    if model.get('decisions'):
        init_vf_list = [init_guesses[dec_input]
                        for dec_input in model['decisions']['inputs']]  # let us for now assume that this must be present
        model['init_vf'] = jnp.array(init_vf_list)

        # check if initial decision functions and the distribution have same shapes
        check_shapes(model['distributions'], model['init_vf'], dist_names)

    # try if function works on initvals
    model['init_run'] = {}
    try:
        check_initial_values(model, shocks, par)
    except:
        if raise_errors:
            raise

    if verbose:
        print("(load:) Parsing done.")

    # add new model to cache
    cached_mdicts += (mdict_raw,)
    cached_models += (model,)

    return model



"""Internal subfunctions for heterogeneous agent models
"""

import jax
import jax.numpy as jnp


def _backwards_stst_cond(carry):
    _, (vf, _, _), (vf_old, cnt), (_, tol, maxit) = carry
    cond0 = jnp.abs(vf - vf_old).max() > tol
    cond1 = cnt < maxit
    return jnp.logical_and(cond0, cond1)


def _backwards_stst_body(carry):
    (x, par), (vf, _, _), (_, cnt), (func, tol, maxit) = carry
    return (x, par), func(x, x, x, x, vf, pars=par), (vf, cnt + 1), (func, tol, maxit)


def backwards_sweep_stst(x, par, carry):
    _, (vf, decisions_output, exog_grid_vars), (_, cnt), _ = jax.lax.while_loop(
        _backwards_stst_cond, _backwards_stst_body, ((x, par), *carry))
    return vf, decisions_output, exog_grid_vars, cnt


def _backwards_step(carry, i):

    vf, X, shocks, func_backw, stst = carry
    vf, decisions_output, exog_grid_vars = func_backw(
        X[:, i], X[:, i+1], X[:, i+2], VFPrime=vf, shocks=shocks[:, i])

    return (vf, X, shocks, func_backw, stst), decisions_output


def backwards_sweep(x, x0, shocks, stst, vfSS, horizon, func_backw):

    X = jnp.hstack((x0, x, stst)).reshape(horizon+1, -1).T

    _, decisions_output_storage = jax.lax.scan(
        _backwards_step, (vfSS, X, shocks, func_backw, stst), jnp.arange(horizon-1), reverse=True)
    decisions_output_storage = jnp.moveaxis(
        decisions_output_storage, 0, -1)

    return decisions_output_storage


def _forwards_step(carry, i):

    dist_old, decisions_output_storage, func_dist = carry
    dist = func_dist(dist_old, decisions_output_storage[..., i])

    return (dist, decisions_output_storage, func_dist), dist_old


def forwards_sweep(decisions_output_storage, dist0, horizon, func_dist):

    _, dists_storage = jax.lax.scan(
        _forwards_step, (dist0, decisions_output_storage, func_dist), jnp.arange(horizon-1))
    dists_storage = jnp.moveaxis(dists_storage, 0, -1)

    return dists_storage


def final_step(x, dists_storage, decisions_output_storage, x0, shocks, stst, horizon, nshpe, pars, func_eqns):

    X = jnp.hstack((x0, x, stst)).reshape(horizon+1, -1).T
    out = func_eqns(X[:, :-2].reshape(nshpe), X[:, 1:-1].reshape(nshpe), X[:, 2:].reshape(
        nshpe), stst, shocks, pars, dists_storage, decisions_output_storage)

    return out


def second_sweep(x, decisions_output_storage, x0, dist0, shocks, forwards_sweep, final_step):

    # forwards step
    dists_storage = forwards_sweep(decisions_output_storage, dist0)
    # final step
    out = final_step(x, dists_storage,
                     decisions_output_storage, x0, shocks)

    return out


def stacked_func_het_agents(x, x0, dist0, shocks, backwards_sweep, second_sweep):

    # backwards step
    decisions_output_storage = backwards_sweep(x, x0, shocks)
    # combined step
    out = second_sweep(x, decisions_output_storage, x0, dist0, shocks)

    return out



# -*- coding: utf-8 -*-

import os
import jax.numpy as jnp
import econpizza as ep

filepath = os.path.dirname(__file__)


def test_bh(create=False):

    mod = ep.load(ep.examples.bh, raise_errors=False)
    _ = mod.solve_stst()

    state = jnp.zeros(len(mod["variables"]))
    state = state.at[:-1].set([0.1, 0.2, 0.0])

    x, flag = ep.find_path_shooting(
        mod, state, horizon=50, max_horizon=500, tol=1e-8, verbose=2)

    path = os.path.join(filepath, "test_storage", "bh.npy")

    assert flag == 0
    if create:
        jnp.save(path, x)
        print(f'Test file updated at {path}')
    else:
        test_x = jnp.load(path)
        assert jnp.allclose(x, test_x)


def test_nk(create=False):

    mod = ep.load(ep.examples.nk)
    _ = mod.solve_stst()

    state = mod["stst"].copy()
    state["beta"] *= 1.02

    x, flag = ep.find_path_shooting(mod, state.values(), horizon=10,
                                    max_horizon=10, verbose=2)

    path = os.path.join(filepath, "test_storage", "nk.npy")

    assert flag == 0
    if create:
        jnp.save(path, x)
        print(f'Test file updated at {path}')
    else:
        test_x = jnp.load(path)
        assert jnp.allclose(x, test_x)


def test_stacked(create=False):

    mod = ep.load(ep.examples.nk)
    _ = mod.solve_stst()

    shk = ("e_beta", 0.02)

    x, flag = ep.find_path_stacking(mod, shock=shk, horizon=50)

    path = os.path.join(filepath, "test_storage", "stacked.npy")

    assert flag == 0
    if create:
        jnp.save(path, x)
        print(f'Test file updated at {path}')
    else:
        test_x = jnp.load(path)
        assert jnp.allclose(x, test_x)


def test_hank(create=False):

    mod_dict = ep.parse(ep.examples.hank)
    mod = ep.load(mod_dict)
    _ = mod.solve_stst(tol=1e-8)

    x0 = mod['stst'].copy()
    x0['beta'] *= 1.01  # setting a shock on the discount factor

    x, flag = mod.find_path(init_state=x0.values(), horizon=50)
    x_lin, _ = mod.find_path_linear(init_state=x0.values(), horizon=50)
    het_vars = mod.get_distributions(x)
    dist = het_vars['dist']

    path_x = os.path.join(filepath, "test_storage", "hank.npy")
    path_x_lin = os.path.join(filepath, "test_storage", "hank_lin.npy")
    path_dist = os.path.join(filepath, "test_storage", "hank_dist.npy")

    assert flag == 0
    if create:
        jnp.save(path_x, x)
        jnp.save(path_x_lin, x_lin)
        jnp.save(path_dist, dist)
        print(f'Test file updated at {path_x},{path_x_lin} and {path_dist}')
    else:
        test_x = jnp.load(path_x)
        test_x_lin = jnp.load(path_x_lin)
        test_dist = jnp.load(path_dist)
        assert jnp.allclose(x, test_x)
        assert jnp.allclose(x_lin, test_x_lin)
        assert jnp.allclose(dist, test_dist)


def test_hank_labor(create=False):

    mod_dict = ep.parse(ep.examples.hank_labor)
    mod = ep.load(mod_dict)
    _ = mod.solve_stst(tol=1e-8)

    shocks = ('e_beta', .005)

    x, flag = mod.find_path(shocks, horizon=50)

    path = os.path.join(filepath, "test_storage", "hank_labor.npy")

    assert flag == 0
    if create:
        jnp.save(path, x)
        print(f'Test file updated at {path}')
    else:
        test_x = jnp.load(path)
        assert jnp.allclose(x, test_x)


def test_hank2(create=False):

    mod_dict = ep.parse(ep.examples.hank2)
    mod = ep.load(mod_dict)
    _ = mod.solve_stst(tol=1e-6)

    x0 = mod['stst'].copy()
    x0['beta'] *= 1.005  # setting a shock on the discount factor

    x, flag = mod.find_path(init_state=x0.values(), horizon=50)

    path = os.path.join(filepath, "test_storage", "hank2.npy")

    assert flag == 0
    if create:
        jnp.save(path, x)
        print(f'Test file updated at {path}')
    else:
        test_x = jnp.load(path)
        assert jnp.allclose(x, test_x)


def test_solid(create=False):

    mod_dict = ep.parse(ep.examples.hank)
    mod = ep.load(mod_dict)
    _ = mod.solve_stst(tol=1e-8)

    shocks = ('e_beta', .01)

    x, flag = mod.find_path(shocks, use_solid_solver=True,
                            horizon=20, chunk_size=90)

    path = os.path.join(filepath, "test_storage", "hank_solid.npy")

    assert flag == 0
    if create:
        jnp.save(path, x)
        print(f'Test file updated at {path}')
    else:
        test_x = jnp.load(path)
        assert jnp.allclose(x, test_x)


def create_all():

    test_bh(create=True)
    test_nk(create=True)
    test_stacked(create=True)
    test_hank(create=True)
    test_hank_labor(create=True)
    test_hank2(create=True)
    test_solid(create=True)



# -*- coding: utf-8 -*-
"""functions for the one-asset HANK model without labor choice. Heavily inspired by https://github.com/shade-econ/sequence-jacobian/#sequence-space-jacobian
"""

import jax
import jax.numpy as jnp
from grgrjax import jax_print
from econpizza.utilities.interp import interpolate


def hh_init(a_grid, skills_grid):
    return jnp.ones((skills_grid.shape[0], a_grid.shape[0]))*1e-2


def hh(Va_p, a_grid, skills_grid, w, n, T, R, beta, sigma_c, sigma_l):
    """A single backward step via EGM
    """

    # MUC as implied by next periods value function
    ux_nextgrid = beta * Va_p

    # consumption can be readily obtained from MUC and MU of labor
    labor_inc = skills_grid[:, None]*n*w
    c_nextgrid = ux_nextgrid**(-1/sigma_c) + labor_inc/(1 + sigma_l)

    # get consumption in grid space
    lhs = c_nextgrid - labor_inc + a_grid[None, :] - T[:, None]
    rhs = R * a_grid

    c = interpolate(lhs, rhs, c_nextgrid)

    # get todays distribution of assets
    a = rhs + labor_inc + T[:, None] - c

    # fix consumption and labor for constrained households
    c = jnp.where(a < a_grid[0], labor_inc + rhs +
                  T[:, None] - a_grid[0], c)
    a = jnp.where(a < a_grid[0], a_grid[0], a)

    # calculate new MUC
    Va = R * (c - labor_inc/(1 + sigma_l)) ** (-sigma_c)

    return Va, a, c


def transfers(skills_stationary, Div, Tax, skills_grid):
    # hardwired incidence rules are proportional to skill; scale does not matter
    rule = skills_grid
    div = Div / jnp.sum(skills_stationary * rule) * rule
    tax = Tax / jnp.sum(skills_stationary * rule) * rule
    T = div - tax
    return T



from setuptools import setup, find_packages
from os import path
from econpizza import __version__

# read the contents of the README file
this_directory = path.abspath(path.dirname(__file__))
with open(path.join(this_directory, "README.rst"), encoding="utf-8") as f:
    long_description = f.read()

setup(
    long_description=long_description,
    long_description_content_type="text/x-rst",
    url="https://github.com/gboehl/econpizza",
    name='econpizza',
    version=__version__,
    author="Gregor Boehl",
    author_email="admin@gregorboehl.com",
    description="Solve nonlinear perfect foresight models with heterogeneous agents",
    license="MIT",
    classifiers=[
        "License :: OSI Approved :: MIT License",
        "Programming Language :: Python :: 3",
        "Programming Language :: Python :: 3.9",
    ],
    include_package_data=True,
    packages=['econpizza', 'econpizza.parser',
              'econpizza.utilities', 'econpizza.solvers'],
    package_data={"econpizza": ["examples/*"]},
    extras_require={
        'linear': ['grgrlib>=0.1.22'],
    },
    install_requires=[
        "jax>=0.3.20",
        "jaxlib",
        "grgrjax",
        "pyyaml",
    ],
)



"""Tools accessible during runtime
"""

import jax
import jax.numpy as jnp
from jax.experimental.host_callback import id_print as jax_print
from .utilities.interp import interpolate


def percentile(x, dist, share):
    """percentiles function
    """
    # reshape inputs flattend for each period t
    horizon = x.shape[-1]
    x_flat = x.reshape(-1, horizon)
    dist_flat = dist.reshape(-1, horizon)

    # distribution sorted according to x
    sorted_args = jnp.argsort(x_flat, axis=0)
    dist_sorted = jnp.take_along_axis(dist_flat, sorted_args, axis=0)
    x_sorted = jnp.take_along_axis(x_flat, sorted_args, axis=0)

    # cummulative sums
    dist_cumsum = jnp.cumsum(dist_sorted, axis=0)
    x_cumsum = jnp.cumsum(x_sorted*dist_sorted, axis=0)

    # interpolate
    return interpolate(dist_cumsum.T, (share,), x_cumsum.T).flatten()/x_cumsum[-1]
